-- NPC dialogue module

require "incl_dialogwindow"

RESPONSE_ACTION = "Responses"
Modules.Dialogue = true

DialogueConfig = {
	RememberPlayers = true, -- If players should have data if they have met this NPC before
	OverrideBaseAI = true -- If AI dialogue should be overridden - BE CAREFUL with this, it can cause side-effects especially if vanilla code changes
}

if AI ~= nil then
	DialogueConfig.RememberPlayers =  AI.GetSetting("RememberPlayers") or DialogueConfig.RememberPlayers
	DialogueConfig.OverrideBaseAI = AI.GetSetting("OverrideVanillaDialog") or DialogueConfig.OverrideBaseAI
end

-- Dialogue data (see below)
Dialogue = {
	Situations = {}, -- See below
	Nodes = {} -- See below
}

--[[ Dialogue table format
Situations = { -- Starting points, in priority order
	{
		Conditions = {}, -- Starting conditions, as condition system table; empty or nil means always
		Node = "NodeName", -- Name of the node to begin with
	}
},

Nodes = {
	Beginning = {
		Text = "NPC's speech",
		Actions = {}, -- Actions to be performed when entering the node

		Responses = {
			{
				Text = "Response", -- Text to display
				Handle = "", -- Node name to show, or special action name; nil or "" to end dialogue
				Conditions = {}, -- Conditions for this option to BE PRESENT
				Actions = {}, -- Actions to do when user selects this response (not implemented)
				PresentActions = {} -- Actions to do when this response is present
			},

			{"Response", "", conditions} -- This is simpler way: [1] is text, [2] is handle; actions are defined in old way
		}
	},

	RedirectNode = { -- Special redirect node
		Redirect = {
			{
				Conditions = {}, -- Conditions, as condition system table
				Node = "" -- Node to use if conditions are met
			}
		}
	}
}
]]

-- Special handles (merchant etc.)
SpecialHandles = {}

function StartDialogue(user)
	--DebugMessage("UseObject")
	
	--- We need this for the quest/Folio System ~Yorlik
	user:SendMessage("NPCInteraction",this)
	---
	
	local entry = GetEntryNode(user)
	if entry == nil then return end

	ShowNode(entry, user)
	if DialogueConfig.RememberPlayers then
		MarkNPCMeet(user)
	end
end

-- Gets the entry node of this dialogue
function GetEntryNode(user)
	if Dialogue.Situations == nil then
		return Dialogue.Nodes.Intro
	end

	for priority,situation in ipairs(Dialogue.Situations) do
		if CheckConditions(situation.Conditions, user) then
			return situation.Node, priority
		end
	end

	return nil
end

function ShowNode(name, user)
	--DebugMessage("Node name is "..tostring(name))
	--DebugMessage("user is (again?) "..tostring(user))
	if name == nil or name == "" then return end
	if SpecialHandles[name] ~= nil then
		SpecialHandles[name](user)
		return
	end

	--DebugMessage("No SpecialHandles")
	local node = Dialogue.Nodes[name]
	if node == nil then return end
	--DebugMessage("Made it past two returns!")

	local redirect = node.Redirect
	if redirect ~= nil then
		for priority,situation in ipairs(redirect) do
			if CheckConditions(situation.Conditions, user) then
				ShowNode(situation.Node, user)
				return
			end
		end
		return
	end

	local responses = {}

	-- Create responses
	for priority,response in ipairs(node.Responses) do
		local conditions = response.Conditions or response[3]
		if CheckConditions(conditions, user) then
			local text = response.Text or response[1]
			if type(text) == "function" then
				text = text(user, name)
			end
			local handle = response.Handle or response[2]
			if type(handle) == "function" then
				handle = handle(user, name)
			end
			if text ~= nil then
				responses[#responses+1] = {text = text, handle = handle}
			end
			PerformActions(response.PresentActions, user)
		end
	end

	-- Perform actions
	PerformActions(node.Actions, user)

	-- Get the text - if it's a function call it
	local rawtext = node.Text
	-- Dialog text preprocess stage 1: function result
	if type(rawtext) == "function" then
		rawtext = rawtext(user, name)
	end

	-- Dialog text preprocess stage 2: 
	-- table ==> In-text parameters in incl_dialogwindow.lua ~Yorlik
	-- the function above could return this table too
	if type(rawtext) == "table" then
		text = rawtext.text
		if type(text)=="function" then text=text(user, name) end
		DebugMessage("1. type(text)"..type(text))
	elseif  type(rawtext) == "string" then
		text = rawtext
		DebugMessage("2. type(text)"..type(text))
	elseif rawtext == nil then 
		DebugMessage("ERROR: Rawtext is nil.")
		return
	else
		DebugMessage("ERROR: Rawtext type error:"..type(rawtext))
		return
	end
	DebugMessage("3. type(text)"..type(text))


	text = text:gsub("\n"," ") -- Line break equals space like in web
	text = text:gsub("<br>", "\n") -- ... and we have tag for true line break
	local displayName = StripColorFromString(user:GetName())
	-- macro processing
	text = text:gsub("{player}", displayName)

	-- Yorliks hack part 2
	if type(rawtext) == "table" then
		sendtext = deepcopy(rawtext)
		sendtext.text=text
	else 
		sendtext = text
	end

	-- Process responses texts
	for k,v in pairs(responses) do 
		local tmp = v.text:gsub("\n"," ") -- Line break equals space like in web
		tmp = tmp:gsub("<br>", "\n") -- ... and we have tag for true line break
		local displayName = StripColorFromString(user:GetName())
		responses[k].text= tmp:gsub("{player}", displayName)
	end

	---- Attempt to give windows unique IDs per node to allow correct xy positioning
	---- It doesn't really work: Window jumping sucks, and it creates new windows when you re-click the NPC.
	-- local unique = tostring(DateTime.UtcNow.Ticks) -- guarantee unique window Id for x-y repositioning
	-- RegisterSingleEventHandler(EventType.DynamicWindowResponse, "Responses"..name, HandleResponse)

	NPCInteraction(sendtext, this, user, "Responses", responses)
	--DebugMessage("Sended node with "..node.Text.." to "..tostring(user))
	--DebugTable(responses)
end

-- Handles responses for player talk
function HandleResponse(user, handle)
	--DebugMessage("Handle response "..handle)
	--DebugMessage("Handle not nil!")
	--DebugMessage("user is "..tostring(user))

	ShowNode(handle, user or {})
end

function MarkNPCMeet(user)
	local meetTable = user:GetObjVar("NPCMeetTable") or {}
	meetTable[this:GetCreationTemplateId()] = true
	user:SetObjVar("NPCMeetTable", meetTable)
end

if not Modules.Talky then -- If yes, use talk data handling
	if initializer ~= nil and initializer.DialogueData ~= nil then
		this:SetObjVar("DialogueData", initializer.DialogueData)
	end

	if this:HasObjVar("DialogueData") then
		AddPostRequire("data_dialogue."..this:GetObjVar("DialogueData"))
	end
end

if Modules.NpcBase then -- Override vanilla dialogue, if it was asked
	AddPostTask(function ()
		if DialogueConfig.OverrideBaseAI then
			OverrideEventHandler("default:base_ai_conversation", EventType.Message, "UseObject", StartDialogue) -- Umm, I guess it is something like this?
		else
			RegisterEventHandler(EventType.Message, "UseObject", StartDialogue)
		end
	end)
else
	RegisterEventHandler(EventType.Message, "UseObject", StartDialogue)
end
RegisterEventHandler(EventType.DynamicWindowResponse, "Responses", HandleResponse)
RegisterEventHandler(EventType.Message, "StartDialogue", StartDialogue)
RegisterEventHandler(EventType.Message, "ShowDialogNode", ShowNode)
RegisterEventHandler(EventType.Message, "MarkNPCMeet", MarkNPCMeet)